#include "ttm.h"

extern int ttm_init_ctl(void);
extern void ttm_uninit_ctl(void);

extern struct net init_net;
int debug_level = 0;
unsigned long sk_data_ready_addr = 0;

MODULE_AUTHOR("sTeeL <steel_mental@gmail.com>");
MODULE_DESCRIPTION("Tunnel Though Module");
MODULE_LICENSE("GPL");
//MODULE_PARM(debug_level, "i");
module_param(debug_level, int, S_IRUGO);
MODULE_PARM_DESC(debug_level, "debug level");

module_param(sk_data_ready_addr, ulong, 0644);
MODULE_PARM_DESC(sk_data_ready_addr, "addr of function sock_def_readable");
//unsigned long sk_data_ready_addr = 0;

struct proc_dir_entry *ttm_dir_entry = NULL, *ttm_stats_entry = NULL;

/*******************************************************
 * Statistics of ttm in proc /proc/sys/ttm/ttm_stats *
 *******************************************************/

struct ttm_stats_entry ttm_stats[] = {
	TTM_STAT_ITEM("recv_sock_ttm", SYN_RECV_SOCK_TTM_CNT),
	TTM_STAT_ITEM("recv_sock_no_ttm", SYN_RECV_SOCK_NO_TTM_CNT),
	TTM_STAT_ITEM("getname_ttm_ok", GETNAME_TTM_OK_CNT),
	TTM_STAT_ITEM("getname_ttm_mismatch", GETNAME_TTM_MISMATCH_CNT),
	TTM_STAT_ITEM("getname_ttm_bypass", GETNAME_TTM_BYPASS_CNT),
	TTM_STAT_ITEM("getname_ttm_empty", GETNAME_TTM_EMPTY_CNT),
	TTM_STAT_ITEM("ip64_address_alloc", IP64_ADDR_ALLOC_CNT),
	TTM_STAT_ITEM("ip64_address_free", IP64_ADDR_FREE_CNT),
	TTM_STAT_END
};

DEFINE_TTM_STAT(struct ttm_stat_mib, ext_stats);

#ifndef NIPQUAD
#define NIPQUAD(addr) \
                ((unsigned char *)&addr)[0], \
                ((unsigned char *)&addr)[1], \
                ((unsigned char *)&addr)[2], \
                ((unsigned char *)&addr)[3]
#endif

#ifndef NIPQUAD_FMT
#define NIPQUAD_FMT "%u.%u.%u.%u"
#endif

#ifdef CONFIG_PROC_FS

static int
ttm_stats_show(struct seq_file *seq, void *v)
{
	int i, j;

	/* print CPU first */
	seq_printf(seq, "                                  ");
	for (i = 0; i < NR_CPUS; i++)
		if (cpu_online(i))
			seq_printf(seq, "CPU%d       ", i);
	seq_putc(seq, '\n');

	i = 0;
	while (NULL != ttm_stats[i].name) {
		seq_printf(seq, "%-25s:", ttm_stats[i].name);
		for (j = 0; j < NR_CPUS; j++)
			if (cpu_online(j)) {
				seq_printf(seq, "%10lu ",
					   *(((unsigned long *) per_cpu_ptr(ext_stats, j)) + ttm_stats[i].entry));
			}
		seq_putc(seq, '\n');
		i++;
	}

	return 0;
}

static int
ttm_stats_seq_open(struct inode *inode, struct file *file)
{
	return single_open(file, ttm_stats_show, NULL);
}

static struct file_operations ttm_stats_fops = {
	.owner = THIS_MODULE,
	.open = ttm_stats_seq_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = single_release,
};
/*
static struct ctl_table ttm_vars[] = {
	{
	 .ctl_name = 0,
	}
};

#define CTL_TTM 10001
static ctl_table ttm_root_table[] = {
	{
	 .ctl_name = CTL_TTM,
	 .procname = "ttm",
	 .mode = 0555,
	 .child = ttm_vars
	},
	{.ctl_name = 0}
};

struct ctl_table_header *sysctl_header;
*/
#endif


/*change register cr0 to make memory writeable*/
static unsigned long clear_and_return_cr0(void)
{
    unsigned long cr0 = 0;
    unsigned long ret;
    asm volatile("movq %%cr0, %%rax"
                     : "=a" (cr0)
    );
    ret = cr0;

    /* clear the 20 bit of CR0, a.k.a WP bit */
    cr0 &= 0xfffeffff;
    asm volatile ("movq %%rax, %%cr0"
                       :: "a" (cr0)
    );
    return ret;
}

/*when write finished, recover the cr0*/
void setback_cr0(unsigned long val)
{
    asm volatile("movq %%rax, %%cr0"
                   :: "a" (val)
    );
}


/* Parse TCP options in skb, try to get client ip, port
 * @param skb [in] received skb, it should be a ack/ack-get packet.
 * @return NULL if we don't get client ip/port;
 *         value of ttm_data in ret_ptr if we get client ip/port.
 */
static void * get_ttm_data(struct sk_buff *skb, int *nat64)
{
	struct tcphdr *th;
	int length;
	unsigned char *ptr;

	struct ttm_data tdata;

	void *ret_ptr = NULL;

	TTM_DBG(12, "get_ttm_data called\n");

	if (NULL != skb) {
		//th = skb->h.th;
		th = tcp_hdr(skb);
		length = (th->doff * 4) - sizeof (struct tcphdr);
		ptr = (unsigned char *) (th + 1);

		while (length > 0) {
			int opcode = *ptr++;
			int opsize;
			switch (opcode) {
			case TCPOPT_EOL:
				return NULL;
			case TCPOPT_NOP:	/* Ref: RFC 793 section 3.1 */
				length--;
				continue;
			default:
				opsize = *ptr++;
				if (opsize < 2)	/* "silly options" */
					return NULL;
				if (opsize > length)
					return NULL;	/* don't parse partial options */
				/* 4to4 case */
				if ((TCPOPT_TTM_CIP == opcode || TCPOPT_FNAT == opcode) &&
                    TCPOLEN_TTM == opsize) {
					memcpy(&tdata, ptr - 2, sizeof (tdata));
					TTM_DBG(12, "find ttm data: ip = %u.%u.%u.%u, port = %u\n", NIPQUAD(tdata.ip),
					ntohs(tdata.port));
					memcpy(&ret_ptr, &tdata, sizeof (ret_ptr));
					TTM_DBG(12, "coded ttm data: %p\n", ret_ptr);
				}
				/* 6to4 case */
				else if (TCPOPT_TTM_CIP_V6 == opcode &&
                    TCPOLEN_TTM_v6 == opsize) {
                    struct ttm_data_v6 *sk_toa_data = 
						kmalloc(sizeof(struct ttm_data_v6), GFP_ATOMIC);
					if (!sk_toa_data) {
						return NULL;
					}
					TTM_INC_STATS(ext_stats, IP64_ADDR_ALLOC_CNT);
                    memcpy(sk_toa_data, ptr - 2, sizeof(struct ttm_data_v6));

					*nat64 = 1;
					return sk_toa_data;
                }
				ptr += opsize - 2;
				length -= opsize;
			}
		}
	}
	return ret_ptr;
}

/* get client ip from socket 
 * @param sock [in] the socket to getpeername() or getsockname()
 * @param uaddr [out] the place to put client ip, port
 * @param uaddr_len [out] lenth of @uaddr
 * @peer [in] if(peer), try to get remote address; if(!peer), try to get local address
 * @return return what the original inet_getname() returns.
 */
int
inet_getname_ttm(struct socket *sock, struct sockaddr *uaddr, int *uaddr_len, int peer)
{
	int retval = 0;
	struct sock *sk = sock->sk;
	struct sockaddr_in *sin = (struct sockaddr_in *) uaddr;
	struct ttm_data tdata;

	TTM_DBG(12, "inet_getname_ttm called, sk->sk_user_data is %p\n", sk->sk_user_data);

	/* call orginal one */
	retval = inet_getname(sock, uaddr, uaddr_len, peer);

	/* set our value if need */
	if (retval == 0 && NULL != sk->sk_user_data && peer) {
		if (sk_data_ready_addr == (unsigned long) sk->sk_data_ready) {
			memcpy(&tdata, &sk->sk_user_data, sizeof (tdata));
			if ((TCPOPT_TTM_CIP == tdata.opcode || TCPOPT_FNAT == tdata.opcode) && TCPOLEN_TTM == tdata.opsize) {
				TTM_INC_STATS(ext_stats, GETNAME_TTM_OK_CNT);
				TTM_DBG(12, "inet_getname_ttm: set new sockaddr, ip %u.%u.%u.%u -> %u.%u.%u.%u, port %u -> %u\n",
						NIPQUAD(sin->sin_addr.s_addr), NIPQUAD(tdata.ip), ntohs(sin->sin_port),
						ntohs(tdata.port));
				sin->sin_port = tdata.port;
				sin->sin_addr.s_addr = tdata.ip;
			} else { /* sk_user_data doesn't belong to us */
				TTM_INC_STATS(ext_stats, GETNAME_TTM_MISMATCH_CNT);
				TTM_DBG(12, "inet_getname_ttm: invalid ttm data, ip %u.%u.%u.%u port %u opcode %u opsize %u\n",
						NIPQUAD(tdata.ip), ntohs(tdata.port), tdata.opcode, tdata.opsize);
			}
		} else {
			TTM_INC_STATS(ext_stats, GETNAME_TTM_BYPASS_CNT);
		}
	} else { /* no need to get client ip */
		TTM_INC_STATS(ext_stats, GETNAME_TTM_EMPTY_CNT);
	}

	return retval;
}

static void 
tcp_6to4_sk_destruct_toa(struct sock *sk) {
        if (sk->sk_user_data) {
                kfree(sk->sk_user_data);
                sk->sk_user_data = NULL;
                TTM_INC_STATS(ext_stats, IP64_ADDR_FREE_CNT);
        }   
        inet_sock_destruct(sk);
}

/* The three way handshake has completed - we got a valid synack -
 * now create the new socket.
 * We need to save ttm data into the new socket.
 * @param sk [out]  the socket
 * @param skb [in] the ack/ack-get packet
 * @param req [in] the open request for this connection
 * @param dst [out] route cache entry
 * @return NULL if fail new socket if succeed.
 */

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9))
static struct sock *
tcp_v4_syn_recv_sock_ttm(struct sock *sk, struct sk_buff *skb, struct open_request *req, struct dst_entry *dst)
#else
static struct sock *
tcp_v4_syn_recv_sock_ttm(struct sock *sk, struct sk_buff *skb, struct request_sock *req, struct dst_entry *dst)
#endif
{
	struct sock *newsock = NULL;
	int nat64 = 0;

	TTM_DBG(12, "tcp_v4_syn_recv_sock_ttm called\n");

	/* call orginal one */
	newsock = tcp_v4_syn_recv_sock(sk, skb, req, dst);

	/* set our value if need */
	if (NULL != newsock && NULL == newsock->sk_user_data) {
		newsock->sk_user_data = get_ttm_data(skb, &nat64);
		if (NULL != newsock->sk_user_data) {
			TTM_INC_STATS(ext_stats, SYN_RECV_SOCK_TTM_CNT);
			if (nat64) {
				newsock->sk_destruct = tcp_6to4_sk_destruct_toa;
			}
		} else {
			TTM_INC_STATS(ext_stats, SYN_RECV_SOCK_NO_TTM_CNT);
		}
		TTM_DBG(12, "tcp_v4_syn_recv_sock_ttm: set sk->sk_user_data to %p\n", newsock->sk_user_data);
	}
	return newsock;
}

/************************
 * Module init and exit *
 ************************/
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9))
extern struct tcp_func ipv4_specific;
#else
extern struct inet_connection_sock_af_ops ipv4_specific;
#endif

/* replace the functions with our functions */
static inline int
hook_ttm_functions(void)
{
	unsigned long cr0;
	struct proto_ops * inet_stream_ops_p = NULL;
	struct inet_connection_sock_af_ops * ipv4_specific_p = NULL;
	/* hook inet_getname */	
	cr0 = clear_and_return_cr0();
        barrier();
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9))
	inet_stream_ops.getname = inet_getname_ttm;
#else
	inet_stream_ops_p = (struct proto_ops *)&inet_stream_ops;
	
	inet_stream_ops_p->getname = inet_getname_ttm;
#endif
	TTM_INFO("CPU [%u] hooked inet_getname <%p> --> <%p>\n", smp_processor_id(), inet_getname,
		 inet_stream_ops.getname);

	/* hook tcp_v4_syn_recv_sock */
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9))
	ipv4_specific.syn_recv_sock = tcp_v4_syn_recv_sock_ttm;
#else
	ipv4_specific_p = &ipv4_specific;
	ipv4_specific_p->syn_recv_sock = tcp_v4_syn_recv_sock_ttm;
#endif
	setback_cr0(cr0);
	TTM_INFO("CPU [%u] hooked tcp_v4_syn_recv_sock <%p> --> <%p>\n", smp_processor_id(), tcp_v4_syn_recv_sock,
		 ipv4_specific.syn_recv_sock);
	return 0;
}

/* replace the functions to original ones */
static int
unhook_ttm_functions(void)
{
	unsigned long cr0;
	struct proto_ops * inet_stream_ops_p = NULL;
	struct inet_connection_sock_af_ops * ipv4_specific_p = NULL;
	cr0 = clear_and_return_cr0();
        barrier();
	/* unhook inet_getname */
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9))
	inet_stream_ops.getname = inet_getname;
#else
	inet_stream_ops_p = (struct proto_ops *)&inet_stream_ops;
	inet_stream_ops_p->getname = inet_getname;
#endif

	TTM_INFO("CPU [%u] unhooked inet_getname\n", smp_processor_id());

	/* unhook tcp_v4_syn_recv_sock */
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9))
	ipv4_specific.syn_recv_sock = tcp_v4_syn_recv_sock;
#else
	ipv4_specific_p = &ipv4_specific;
	ipv4_specific_p->syn_recv_sock = tcp_v4_syn_recv_sock;
#endif
	setback_cr0(cr0);
	TTM_INFO("CPU [%u] unhooked tcp_v4_syn_recv_sock\n", smp_processor_id());
	return 0;
}

/* module init */
static int
ttm_init(void)
{
	//int ret = 0;

	TTM_INFO("TTM " TTM_VERSION " by sTeeL<steel.mental@gmail.com>\n");

	/* alloc statistics array for ttm */
	if (NULL == (ext_stats = alloc_percpu(struct ttm_stat_mib)))
		goto proc_err;

	/* create proc entries */
	//sysctl_header = register_sysctl_table(ttm_root_table, 0);
	/*
	sysctl_header = register_sysctl_table(ttm_root_table);
	if (NULL == sysctl_header) {
		TTM_INFO("cannot register sysctl, /proc/sys/ttm.\n");
		goto proc_err;
	}
	*/

	//ttm_stats_entry = proc_ttm_fops_create("ttm_stats", 0, &ttm_stats_fops);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0))
	ttm_stats_entry = proc_net_fops_create(&init_net, "ttm_stats", 0, &ttm_stats_fops);
#else
    ttm_stats_entry = proc_create("ttm_stats", 0, NULL, &ttm_stats_fops);
#endif
	if (NULL == ttm_stats_entry) {
		TTM_INFO("cannot register ttm_stats.\n");
		goto proc_err;
	}

	/* get the address of function sock_def_readable
	 * so later we can know whether the sock is for rpc, tux or others */
	#if 0
	if (sk_data_ready_addr == 0) {
		sk_data_ready_addr = kallsyms_lookup_name("sock_def_readable");
		TTM_INFO("CPU [%u] sk_data_ready_addr = kallsyms_lookup_name(sock_def_readable) = %p\n", 
		 	smp_processor_id(), (void*)sk_data_ready_addr);
		if(0 == sk_data_ready_addr) {
			TTM_INFO("cannot find sock_def_readable.\n");
			goto proc_err;
		}
	}
	#endif
	if (sk_data_ready_addr == 0) {
		TTM_INFO("cannot find sock_def_readable.\n");
           	goto proc_err;
	}

    if (ttm_init_ctl() < 0) {
        goto proc_err;
    }

	TTM_INFO("sk_data_ready_addr %lu.\n", sk_data_ready_addr);
	hook_ttm_functions();
	TTM_INFO("ttm loaded\n");
	
	return 0;

proc_err:
	if (NULL != ttm_stats_entry) {
		//proc_ttm_remove("ttm_stats");
#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0))
		proc_net_remove(&init_net, "ttm_stats");
#else
		proc_remove(ttm_stats_entry);
#endif
		ttm_stats_entry = NULL;
	}
	/*
	if (NULL != sysctl_header) {
		unregister_sysctl_table(sysctl_header);
		sysctl_header = NULL;
	}
	*/
	if (NULL != ext_stats) {
		free_percpu(ext_stats);
		ext_stats = NULL;
	}
	return -1;
}

/* module cleanup*/
static void
ttm_exit(void)
{
    ttm_uninit_ctl();
	unhook_ttm_functions();
	synchronize_net();

	/* remove proc entries */
	if (NULL != ttm_stats_entry) {
		//proc_ttm_remove("ttm_stats");

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0))
		proc_net_remove(&init_net, "ttm_stats");
#else
        proc_remove(ttm_stats_entry);
#endif

		ttm_stats_entry = NULL;
	}
	/*
	if (NULL != sysctl_header) {
		unregister_sysctl_table(sysctl_header);
		sysctl_header = NULL;
	}
	*/
	if (NULL != ext_stats) {
		free_percpu(ext_stats);
		ext_stats = NULL;
	}
	TTM_INFO("ttm unloaded\n");
}

module_init(ttm_init);
module_exit(ttm_exit);
